import torch
from tqdm import tqdm
import torchvision.utils as tvu
import os
import numpy as np

def compute_alpha(beta, t):
    beta = torch.cat([torch.zeros(1).to(beta.device), beta], dim=0)
    a = (1 - beta).cumprod(dim=0).index_select(0, t + 1).view(-1, 1, 1, 1)
    return a


# x: 初始噪声；seq: 时间下标序列；model: 扩散模型；betas: beta序列；H_funcs: 观测矩阵；y_0: 观测；sigma_0: 观测噪声标准差；clas_fn: 分类器；classes: 类别
def efficient_generalized_steps(x, seq, model, b, H_funcs, y_0, sigma_0, etaB, etaA, etaC, cls_fn=None, classes=None):
    with torch.no_grad():
        #initialize x_T as given in the paper
        largest_alphas = compute_alpha(b, (torch.ones(x.size(0)) * seq[-1]).to(x.device).long())
        
        #setup iteration variables
        # x = H_funcs.V(init_y.view(x.size(0), -1)).view(*x.size())
        y_upsampling = H_funcs.upsampling(y_0)
        n = x.size(0)
        seq_next = [-1] + list(seq[:-1])
        x0_preds = []
        xs = [x]


        t = (torch.ones(n) * seq[-1]).to(x.device)
        # alpha_t和alpha_{t-1}
        at = compute_alpha(b, t.long())
        noise = torch.randn_like(x)
        x_T = noise * (1 - at).sqrt()
        et = model(x_T, t)
        if et.size(1) == 6:
            et = et[:, :3]
        x0_t = (x_T - et * (1 - at).sqrt()) / at.sqrt()

        
        # print(x0_t)
        # print(y_upsampling)
        v = None
        beta=0.0
        # 1000steps: 0.1
        # 100steps: 1.8
        # 20steps: 1.8
        lr=1.0
        N=1

        lam=0.8
        init_noise = torch.randn_like(x0_t)
        #iterate over the timesteps
        for i, j in tqdm(zip(reversed(seq), reversed(seq_next))):
            for _ in range(N):
                if 1 - at[0,0,0,0] > 16 * sigma_0**2:
                    t = (torch.ones(n) * i).to(x.device)
                    next_t = (torch.ones(n) * j).to(x.device)
                    # alpha_t和alpha_{t-1}
                    at = compute_alpha(b, t.long())
                    at_next = compute_alpha(b, next_t.long())
                    # 重新加噪，Stochastic Encoding
                    xt = (at - 16 * sigma_0**2).sqrt() * x0_t + torch.randn_like(x0_t) * (1 - at + 16 * sigma_0**2).sqrt()
                    # 重新加噪，Restricted Resampling
                    # xt = at.sqrt() * x0_t + (1-at_next).sqrt() * init_noise + torch.randn_like(x0_t) * (at_next - at).sqrt()
                    # 计算error。是一个epsilon-predictor
                    if cls_fn == None:
                        et = model(xt, t)
                    else:
                        et = model(xt, t, classes)
                        et = et[:, :3]
                        et = et - (1 - at).sqrt()[0,0,0,0] * cls_fn(x,t,classes)
                    
                    if et.size(1) == 6:
                        et = et[:, :3]
                    
                    # 计算新的x0_t
                    at_part = at / at_next
                    x0_t_new = (xt - et * (1 - at_part).sqrt()) / at_part.sqrt()

                    # 扩散模型的更新方向
                    diff = x0_t_new - x0_t
                    # 观测的梯度方向
                    gradient_obs = H_funcs.upsampling(H_funcs.downsampling(x0_t) - y_0)
                    # 合方向
                    d = diff - lam * gradient_obs
                    if v is None:
                        v = d
                    else:
                        v = beta * v + (1-beta) * d
                    x0_t += lr * v
                    xt_next = x0_t
                    x0_preds.append(x0_t.to('cpu'))
                    xs.append(xt_next.to('cpu'))

    return xs, x0_preds